#include "las2.c"
#include "svdutil.c"
#include "svdlib.c"
#include "main.h"

/*********************************************************/
/****** Functions to compute the flattened matrix ********/
/*********************************************************/


/************************************************************************************/
/* Compare binary vectors encoding patterns, using a mask for split */
/* This has both row and column versions*/

int bin_compC(const void *n1, const void *n2)
{  
  int i;
  unsigned int a,b;

  const unsigned int *num1 =(const unsigned int *)n1;
  const unsigned int *num2 =(const unsigned int *)n2;

  for (i=0; i<nints; i++)
    {
      a=(num1[i] & split_taxaC[i]);
      b=(num2[i] & split_taxaC[i]);
      if ( a<b ) return -1;
      else if ( a>b ) return  1;
    }
  return 0;
}


int bin_compR(const void *n1, const void *n2)
{
  int i;
  unsigned int a,b;

  const unsigned int *num1 =(const unsigned int *)n1;
  const unsigned int *num2 =(const unsigned int *)n2;

  for (i=0; i<nints; i++)
    {
      a=(num1[i] & split_taxaR[i]);
      b=(num2[i] & split_taxaR[i]);
      if ( a<b ) return -1;
      else if ( a>b ) return  1;
    }
  return 0;
}


/*****************************************************************************************/

/** Create binary encoding of patterns for use in flattening **/
int Data_Bin()
{

  /*Note: we remove any sites with any characters other than ATCG*/

  int  badBase, i, j, jj, k, m;
  unsigned int *pArr_u_bin;

  /*number of bases we can encode in one 32-bit integer*/
  nfit = 32/2; 
  /*number of integers needed for encoding patterns 
                           (plus at least 2 extra bits)*/
  nints = (nseq/nfit) + 1; 


  /*Allocate array for unique site patterns*/
  /*Note: we allocate as contiguous block for use in qsort*/
  /*      Sorting will also use 2 extra ints in each column*/
  pArr_u_bin = (unsigned int*)malloc(num_unique*(nints+2)*sizeof(unsigned int)
				     );
  if (pArr_u_bin==NULL)
    {
      printf("Can't memalloc pArr_u_bin\n");
      return 0;
    }
  /*now allocate for pointers to patterns(rows of array)*/
  ppBase_u_bin=malloc(num_unique*sizeof(unsigned int*));
  if (ppBase_u_bin==NULL)
    {
      printf("Can't memalloc ppBase_u_bin\n");
      return 0;
    }
  /*complete allocation by pointing ppBase_u_bin to pArr_u_bin*/
  for (jj=0; jj<num_unique; jj++)
    {
      ppBase_u_bin[jj]=pArr_u_bin+jj*(nints+2);
    }


  /*Encode unique patterns with ONLY ATCG in binary format*/
  jj=0; //initialize pattern number in ppBase_u_bin                                                         
  for (j=0; j<num_unique; j++) //loop on unique sites                                                       
    {
      badBase=0; /*initialize flag,  1 will indicate non-ATCG*/
      for (i=0; i<nseq; i++)  /*loop on sequences*/
        {
          if (ppBase_unique[i][j]<4) /* make sure we have a ATCG*/
            {
              k=i/nfit; //compute word to cose this taxon in                                                
              m=i-k*nfit;//   and location in word, sort of                                                 
              if ( m == 0 ) /*  if beginning new word*/
                {
                  ppBase_u_bin[jj][k]=ppBase_unique[i][j];/*...just store*/
                }
              else
                {
                  ppBase_u_bin[jj][k] ^= ( ppBase_unique[i][j] << (2*m) ); /*... otherwise, shift base and \
									     combine*/
                }
            }
          else badBase=1; // a non-ATCG popped up --- this site will be ignored eventually                  
        }

      if ( badBase==0 ) // if this site had only ATCG                                                       
        {
          ppBase_u_bin[jj][nints-1] ^= (1 << 30);//set high bit so all patterns giveso non-zero encoding    
          ppBase_u_bin[jj][nints]=site_counter[j];// copy counts                                            
          jj++; //don't overwrite binary encoded pattern                                                    
        }
    }
  num_unique_bin=jj;// number of unique patterns in binary encoding                                         

}


// ESA March 2010

/*** Read splits of interest from file, compute svd and scores ***/


float getSVDscore(unsigned int *split_taxa, int sze)
{  
  int j, jj;
  int num_unique_rows, num_unique_cols;
  long int* firstnz; 

  // ESA temporary variables for development
  double score1=-44.4;
  // ESA
  double er=-44.4;
  double ar=-44.4;

  SMat S;
  SVDRec R;

  /*allocate space for binary encoding of splits*/  
  split_taxaR = (unsigned int*)malloc((nints)*sizeof(unsigned int));
  split_taxaC = (unsigned int*)malloc((nints)*sizeof(unsigned int));

  /* initialize split_taxa vector */
  for (j=0; j<nints; j++) {
    split_taxaR[j] = 0;
    split_taxaC[j] = 0;
  }

  //  for (j=0; j<nints; j++) split_taxaR[j] = qq->split_taxa[j]; /*transfer split stored at node of interest*/
  for (j=0; j<nints; j++) split_taxaR[j] = split_taxa[j]; /*transfer split stored at node of interest*/
  for (j=0; j<nints; j++) split_taxaC[j] =~ split_taxaR[j]; /*complementary mask*/

  // printf("for node %d, splitR is %u and splitC is %u\n",qq->number,split_taxaR[0],split_taxaC[0]);
  /*sort on rows, to get row indices, then on columns, to get sparse encoding of flattening*/

  /*sort for rows*/
  qsort(*ppBase_u_bin, num_unique_bin, (nints+2)*sizeof(unsigned int), bin_compR);
  
  /*Enter row numbers into array*/
  jj=0;//initialize row number
  ppBase_u_bin[0][nints+1]=0; // this is always in first row
  for (j=1; j<num_unique_bin; j++)
    {	   	  
      if (bin_compR(ppBase_u_bin[j],ppBase_u_bin[j-1])==1) jj++;
      ppBase_u_bin[j][nints+1]=jj; // record row number
    }
  num_unique_rows=jj+1;

  /*sort for columns*/
  qsort(*ppBase_u_bin, num_unique_bin, (nints+2)*sizeof(unsigned int), bin_compC);
  
  /*allocate space for sparse matrix encoding*/
  long int* rowindex = (long int*) calloc(num_unique_bin,sizeof(long int));
  double* values  = (double*) calloc(num_unique_bin,sizeof(double));

  /* create vector encoding column numbers */
  jj=0; // first column number
  values[0]=(double)ppBase_u_bin[0][nints]/(double)nsite;// copy value of first non-zero for SVD
  
  rowindex[0]=ppBase_u_bin[0][nints+1];/*copy row index of first entry for SVD*/
  ppBase_u_bin[0][nints+1]=0; /*first entry is always in first column*/
  /*Note: we're reusing this space to store column info as we remove row info*/
  for (j=1; j<num_unique_bin; j++)
    { 
      values[j]=(double)(ppBase_u_bin[j][nints])/(double)nsite;/* copy non-zeros for SVD */
      rowindex[j]=ppBase_u_bin[j][nints+1];/* copy row indices for SVD*/
      if (bin_compC(ppBase_u_bin[j],ppBase_u_bin[j-1])==1)/* if new column...*/
	{ 
          jj++;      
	  ppBase_u_bin[jj][nints+1]=j;/*save column coding here temporarily*/
	}
    }
  num_unique_cols=jj+1;

  /*allocate final space for sparse matrix encoding*/
  firstnz = (long int*) calloc( num_unique_cols+1 , sizeof(long int));

  for (j=0; j<num_unique_cols; j++)
    {
      firstnz[j]=ppBase_u_bin[j][nints+1];
    }

  /* copy over column coding for SVD */
  firstnz[j]=num_unique_bin;/* sparse format requires this too*/
  
  /* Create Smat object to use in SVD routines */
  S = svdNewSMat(num_unique_rows,num_unique_cols,num_unique_bin);
  R = svdNewSVDRec();
  
  S->pointr = firstnz;
  S->rowind = rowindex;
  S->value = values;
  
  /* Compute SVD */
  extern long SVDVerbosity;
  SVDVerbosity=0; // print nothing from SVD routine
  //SVDVerbosity=1;        

  R = svdLAS2A(S,16);

  //  printf("the singular values are %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f\n",R->S[0],R->S[1],R->S[2],R->S[3],R->S[4],R->S[5],R->S[6],R->S[7],R->S[8],R->S[9],R->S[10],R->S[11],R->S[12],R->S[13],R->S[14],R->S[15]);
  
  //Below is old score stuff from project with E&J
  // begin score computations
  // Warning: fnorm is actually fnorm^2
  //
  // ESA save current scoring mechanism to score1
  //  score1 = (pow(R->S[0],2)+pow(R->S[1],2)+pow(R->S[2],2)+pow(R->S[3],2))/fnorm;
  //score1 = 1-sqrt((pow(R->S[0],2)+pow(R->S[1],2)+pow(R->S[2],2)+pow(R->S[3],2))/fnorm);
  //er = sqrt(1-(pow(R->S[0],2)+pow(R->S[1],2)+pow(R->S[2],2)+pow(R->S[3],2))/fnorm);
  //ar = er/sqrt((1-pow(4,(1-sze)))*(1-pow(4,(1-(nseq-sze)))));

  score1 = sqrt(pow(R->S[10],2) + pow(R->S[11],2) + pow(R->S[12],2) + pow(R->S[13],2) + pow(R->S[14],2) + pow(R->S[15],2));

  // ESA
  fprintf(scoresfile, "%1.14f ", score1);
  // printf("sze %d nseq %d   score (old er ar) are %f %f %f \n\n",sze, nseq, score1, er, ar);

  fflush(0);
  
  free(firstnz);    
  free(rowindex);   
  free(values);
  
  S->pointr = NULL;
  S->rowind = NULL;
  S->value = NULL;
  
  svdFreeSMat(S);
  svdFreeSVDRec(R);
  
  num_unique_rows=0;
  num_unique_cols=0;

  return(score1);
  
}
