/***************************************************************/
/********** Get scores for splits.                    **********/
/********** Written by L. S. Kubatko, E. S. Allman    **********/
/********** and J. Rhodes, March 2010                 **********/
/********** Modified by LSK July 2012                 **********/
/***************************************************************/

#include "main.h"

/************************/
/*** global variables ***/
/************************/

/*** general ***/
naym *psname;
int nsite, nseq, num_unique, print_data_ind=0, print_data_phylip=0, include_gaps;
int **ppBase, **ppBase_unique, *site_counter;
double fnorm;

/*** file variables for file containing the splits and output file called scoresfile ***/
FILE *splits;
FILE *scoresfile;

/*** variables for binary encoding ***/
int nints, nfit, num_unique_bin;
unsigned int **ppBase_u_bin, *counts, *split_taxaR, *split_taxaC;


/*** Open data input and output files***/

int FileSpecInit(FILESPEC* pFileSpec) {

  // open data file in phylip format

  if ((pFileSpec->infile=fopen("infile","r"))==NULL)
    {
      printf("Can't open infile\n");
      return 0;
    }

  return 1;
}


/*** Memory allocation for all global variables which are pointers ***/

int MemAlloc() {

  int i;

  ppBase = (int**)malloc(nseq*sizeof(int*));
  if (ppBase==NULL)
    {
      printf("Can't memalloc ppBase\n");
      return 0;
    }
  
  for (i=0; i<nseq; i++)
    {
      ppBase[i]=(int*)malloc((nsite+1)*sizeof(int));
      if (ppBase[i]==NULL)
	{
	  printf("Can't memalloc ppBase[%d]\n",i);
	  return 0;
	}
    }
 
  ppBase_unique = (int**)malloc(nseq*sizeof(int*));
  if (ppBase_unique==NULL)
    {
      printf("Can't memalloc ppBase_unique\n");
      return 0;
    }
  
  for (i=0; i<nseq; i++)
    {
      ppBase_unique[i]=(int*)malloc((nsite+5)*sizeof(int));
      if (ppBase_unique[i]==NULL)
	{
	  printf("Can't memalloc ppBase_unique[%d]\n",i);
	  return 0;
	}
    }

  site_counter = (int*)malloc((nsite+1)*sizeof(int));
  if (site_counter==NULL)
    {
      printf("Can't memalloc site_counter\n");
      return 0;
    }
    
  return 1;
}


/*** Transform character sequences (A,G,C,T) ****/
/*** into numerical values (0,1,2,3)         ****/

void transfer(char **cbase,int **nbase,int pres_loc, int trans_length) {

  int i, j;

  for (i=0; i<nseq; i++)  {

    for (j=0; j<trans_length; j++) {

      switch (cbase[i][j]) {
	    
      case 'A': case 'a': nbase[i][j+pres_loc] = 0;  break;
      case 'G': case 'g': nbase[i][j+pres_loc] = 1;  break;
      case 'C': case 'c': nbase[i][j+pres_loc] = 2;  break;
      case 'T': case 't': case 'U': case 'u': nbase[i][j+pres_loc] = 3;  break;
      case 'M': case 'm': nbase[i][j+pres_loc] = 5;  break;
      case 'R': case 'r': nbase[i][j+pres_loc] = 6;  break;
      case 'W': case 'w': nbase[i][j+pres_loc] = 7;  break;
      case 'S': case 's': nbase[i][j+pres_loc] = 8;  break;
      case 'Y': case 'y': nbase[i][j+pres_loc] = 9;  break;
      case 'K': case 'k': nbase[i][j+pres_loc] = 10;  break;
      case 'B': case 'b': nbase[i][j+pres_loc] = 11;  break;
      case 'D': case 'd': nbase[i][j+pres_loc] = 12;  break;
      case 'H': case 'h': nbase[i][j+pres_loc] = 13;  break;
      case 'V': case 'v': nbase[i][j+pres_loc] = 14;  break;
      case 'N': case 'n': nbase[i][j+pres_loc] = 15;  break;
      case '.':nbase[i][j+pres_loc] = nbase[0][j+pres_loc]; break;
      case '-':nbase[i][j+pres_loc]=4; break;
      case '?':nbase[i][j+pres_loc]=15; break;
      default: break;

      }
	  
    }
  }
}


/*** Input the number of sequences (nseq), number of sites (nsite)   ****/
/*** and DNA sequences of taxa from datafile.                        ****/
/*** Then transform DNA sequences into numerical sequences (ppBase)  ****/

naym* InputInit(FILESPEC *pFileSpec) {
  
  int i, k;
  char **ppDNASeq;
  int charLen;
  
  charLen=(nsite+1)*sizeof(char);
  ppDNASeq=(char**)malloc(nseq*sizeof(char*));
  if(ppDNASeq==NULL){
    return NULL;
  }
  
  for (i=0; i<nseq; i++)
    {
      ppDNASeq[i]=(char*)malloc(charLen);
      if(ppDNASeq[i]==NULL){
	return NULL;
      }
    }
  
  psname = (naym*)malloc(nseq*sizeof(naym));
  
  if(psname==NULL){
    return NULL;
  }
  
  for (i=0; i<nseq; i++)
    {
      fscanf(pFileSpec->infile, "%s %s", psname[i],ppDNASeq[i]);

      for (k=0; k<11; k++) if (psname[i][k] == 0) psname[i][k] = ' ';

    }
  
  transfer(ppDNASeq,ppBase,0,nsite);
  
  fclose(pFileSpec->infile);
  pFileSpec->infile=NULL;
  
  for (i=0; i<nseq; i++) {
    free(ppDNASeq[i]);
  }
  
  free(ppDNASeq);
  
  return psname;
  
}


/*** Search all unique sites and count them - counts  are   ***/
/*** stored in the array site_counter.                      ***/


/*** This version also eliminates sites with ambiguity codes ***/

void unique_sites(FILESPEC fspec){

  int i,j,k,s,m=0;
  int num_no_gaps=0;
  int start;

  start=0;

  /* Find the first site with no gaps if include_gaps==0 */

  if (include_gaps==0) {

    for (j=0; j<nsite; j++) {      
      i=0;
      /*while (i<nseq && ppBase[i][j]!=4) { i++; }*/
      while (i<nseq && ppBase[i][j]!=4 && ppBase[i][j]!=15) { i++; }
      
      if (i==nseq) {
	start=j;
	break;
      }
      
    }

  }

  /* First remove sites with all gaps if include_gaps==1 */

  if (include_gaps==1) {

    for (j=0; j<nsite; j++) {
      i=0; 
      while (i<nseq && ppBase[i][j]==4) { i++; }
      if (i==nseq) ppBase[0][j] = 99;
    }
  }

  for (j=start; j<nsite; j++) {
    if ((ppBase[0][j]!=99 && include_gaps==1) || (ppBase[0][j]!=99 && include_gaps==0 && ppBase[0][j]!=4 && ppBase[0][j]!=15)) {
      site_counter[m]=1;
      for (i=0; i<nseq; i++) {
	ppBase_unique[i][m] = ppBase[i][j];
      }
      ppBase[0][j]=99;

      for (k=j+1; k<nsite; k++) {
	i=0;
	while (i< nseq && ppBase_unique[i][m] == ppBase[i][k]) {
	  i++;
	}

	if (i==nseq) {
	  site_counter[m]++;
	  ppBase[0][k]=99;
	}

	if (i<nseq && ppBase[i][k]>=4 && include_gaps==0) {
	  ppBase[0][k]=99;
	}

	if (i<nseq && include_gaps==0) {
	  for (s=0; s<nseq; s++) {
	    if (ppBase[s][k]>=4) ppBase[0][k]=99;
	  }
	}
      }
      m++;
    }
  }

  num_unique=m;

  for (i=0; i<num_unique; i++) num_no_gaps += site_counter[i];

}



/*************   The Main Program  *********************/


int main(){

  FILESPEC filespec;
  naym *sname;
  int i, jj;
  float scorej, min_ind, min_score;

  include_gaps=0;
  fnorm = 0.0;
  
  /*** Read data from files ***/
  if (!FileSpecInit(&filespec)) {
    printf("Can't open needed file \n");
    exit(1);
  }
  fscanf(filespec.infile, "%d %d", &nseq, &nsite);

  /*** Allocate memory ***/
  MemAlloc();

  /*** Call the function InputInit which will read sequences  ***/
  /*** from infile and convert them to numeric using the      ***/
  /*** function transform.  Returns NULL if a problem occurs. ***/
  sname = InputInit(&filespec);
  if (sname==NULL) {
    printf("Can't allocate memory in InputInit\n");
    exit(1);
  }

  /* compute quantities needed to store splits in binary format */
  nfit = 32/2;             /*number of bases we can encode in one 32-bit integer*/
  nints = (nseq/nfit) + 1; /*number of integers needed for encoding patterns (plus at least 2 extra bits)*/


  /***  Find all unique site patterns and store them in ppBase_unique  ***/
  unique_sites(filespec);

  fflush(0);

  /*** Get sum of squared counts, i.e. the square of the Froebenius norm ***/
  for (i=0; i<num_unique; i++) fnorm = fnorm + pow((double)site_counter[i]/(double)nsite,2);

  Data_Bin();

  // open files for reading splits and for writing scores
     
  if ((splits=fopen("splits","r"))==NULL)
    {
      printf("Can't open splits\n");
      return 0;
    }
     
  if ((scoresfile=fopen("scoresfile","w"))==NULL)
    {
      printf("Can't open scoresfile\n");
      return 0;
    }

  int nsplits, split_size, taxon;
  int k, kk;
  int jsplit_num;
  unsigned int *split_taxa;

  split_taxa = (unsigned int*)malloc((nints)*sizeof(unsigned int));

  fscanf(splits, "%d", &nsplits);
  fprintf(scoresfile, "%d\n", nsplits);

  // loop on number splits in file
  min_score = 100000.0;
  min_ind=0;
  for (jsplit_num=0; jsplit_num<nsplits; jsplit_num++)
    {
      // get size of split from file
      fscanf(splits, "%d", &split_size);
      fprintf(scoresfile, "%d ", split_size);

      // zero out contents of split_taxa 
      for (jj=0; jj<nints; jj++) split_taxa[jj] = 0;

      // get taxa in split from file and create mask for split_taxa
      for (i = 0; i < split_size; i++)
	{
	  fscanf(splits, "%d\n", &taxon);
	  // printf("%d ",taxon);

	  kk=(taxon-1)/nfit;
	  k=taxon-1-kk*nfit;
 
	  /*flip 2 bits for mask for this taxon*/
	  split_taxa[kk] ^= (3 << 2*k);

	}

      // get SVD score for current split
      scorej = getSVDscore(split_taxa, split_size);
      //printf("scorejh is %1.14f\n",scorej);
      if (scorej<min_score) {
	min_score = scorej;
	min_ind = jsplit_num;
      }
      if (jsplit_num==0) fprintf(scoresfile," ((A,B),(C,D));\n");
      else if (jsplit_num==1) fprintf(scoresfile," ((A,C),(B,D));\n");
      else if (jsplit_num==2) fprintf(scoresfile," ((A,D),(B,C));\n");

    }

  //  printf("%1.14f ",min_score);

  if (min_ind==0) {//printf("((A,B),(C,D));\n");
    for (k=0; k<2; k++) printf("%c",psname[0][k]);
    printf(",");
    for (k=0; k<2; k++) printf("%c",psname[1][k]);
    printf("|");
    for (k=0; k<2; k++) printf("%c",psname[2][k]);                                                                                                                  
    printf(",");                                                                                                                                                    
    for (k=0; k<2; k++) printf("%c",psname[3][k]);  
    printf("\n"); 
  }

  else if (min_ind==1) { //printf("((A,C),(B,D));\n");
    for (k=0; k<2; k++) printf("%c",psname[0][k]);                                                                                                                  
    printf(",");                                                                                                                                                    
    for (k=0; k<2; k++) printf("%c",psname[2][k]);                                                                                                                  
    printf("|");                                                                                                                                                   
    for (k=0; k<2; k++) printf("%c",psname[1][k]);                                                                                                                  
    printf(",");                                                                                                                                                    
    for (k=0; k<2; k++) printf("%c",psname[3][k]); 
    printf("\n"); 
  }

  else if (min_ind==2) {//printf("((A,D),(B,C));\n");
    for (k=0; k<2; k++) printf("%c",psname[0][k]);                                                                                                                  
    printf(",");                                                                                                                                                    
    for (k=0; k<2; k++) printf("%c",psname[3][k]);                                                                                                                  
    printf("|");                                                                                                                                                   
    for (k=0; k<2; k++) printf("%c",psname[2][k]);                                                                                                                  
    printf(",");                                                                                                                                                    
    for (k=0; k<2; k++) printf("%c",psname[1][k]); 
    printf("\n"); 
  }


  fclose(splits);
  fclose(scoresfile);

   
}


  



