/********************************************************************************/
/********** Code to generate either all quartets or a sample of quartets ********/
/********************************************************************************/

#include "main.h"

/************************/
/*** global variables ***/
/************************/

/*** general ***/
naym *psname;
int nsite, nseq, num_unique, print_data_ind=0, print_data_phylip=0, include_gaps;
int **ppBase, **ppBase_unique, *site_counter;
double fnorm;

/*** file variables for file containing the splits and output file called scoresfile ***/
FILE *splits;
FILE *scoresfile;

/*** variables for binary encoding ***/
int nints, nfit, num_unique_bin;
unsigned int **ppBase_u_bin, *counts, *split_taxaR, *split_taxaC;


/*** Open data input and output files***/

int FileSpecInit(FILESPEC* pFileSpec) {

  // open data file in phylip format

  if ((pFileSpec->infile=fopen("data.phy","r"))==NULL)
    {
      printf("Can't open data.phy\n");
      return 0;
    }

  return 1;
}


/*** Memory allocation for all global variables which are pointers ***/

int MemAlloc() {

  int i;

  ppBase = (int**)malloc(nseq*sizeof(int*));
  if (ppBase==NULL)
    {
      printf("Can't memalloc ppBase\n");
      return 0;
    }
  
  for (i=0; i<nseq; i++)
    {
      ppBase[i]=(int*)malloc((nsite+1)*sizeof(int));
      if (ppBase[i]==NULL)
	{
	  printf("Can't memalloc ppBase[%d]\n",i);
	  return 0;
	}
    }
 
  ppBase_unique = (int**)malloc(nseq*sizeof(int*));
  if (ppBase_unique==NULL)
    {
      printf("Can't memalloc ppBase_unique\n");
      return 0;
    }
  
  for (i=0; i<nseq; i++)
    {
      ppBase_unique[i]=(int*)malloc((nsite+5)*sizeof(int));
      if (ppBase_unique[i]==NULL)
	{
	  printf("Can't memalloc ppBase_unique[%d]\n",i);
	  return 0;
	}
    }

  site_counter = (int*)malloc((nsite+1)*sizeof(int));
  if (site_counter==NULL)
    {
      printf("Can't memalloc site_counter\n");
      return 0;
    }
    
  return 1;
}


/*** Transform character sequences (A,G,C,T) ****/
/*** into numerical values (0,1,2,3)         ****/

void transfer(char **cbase,int **nbase,int pres_loc, int trans_length) {

  int i, j;

  for (i=0; i<nseq; i++)  {

    for (j=0; j<trans_length; j++) {

      switch (cbase[i][j]) {
	    
      case 'A': case 'a': nbase[i][j+pres_loc] = 0;  break;
      case 'G': case 'g': nbase[i][j+pres_loc] = 1;  break;
      case 'C': case 'c': nbase[i][j+pres_loc] = 2;  break;
      case 'T': case 't': case 'U': case 'u': nbase[i][j+pres_loc] = 3;  break;
      case 'M': case 'm': nbase[i][j+pres_loc] = 5;  break;
      case 'R': case 'r': nbase[i][j+pres_loc] = 6;  break;
      case 'W': case 'w': nbase[i][j+pres_loc] = 7;  break;
      case 'S': case 's': nbase[i][j+pres_loc] = 8;  break;
      case 'Y': case 'y': nbase[i][j+pres_loc] = 9;  break;
      case 'K': case 'k': nbase[i][j+pres_loc] = 10;  break;
      case 'B': case 'b': nbase[i][j+pres_loc] = 11;  break;
      case 'D': case 'd': nbase[i][j+pres_loc] = 12;  break;
      case 'H': case 'h': nbase[i][j+pres_loc] = 13;  break;
      case 'V': case 'v': nbase[i][j+pres_loc] = 14;  break;
      case 'N': case 'n': nbase[i][j+pres_loc] = 15;  break;
      case '.':nbase[i][j+pres_loc] = nbase[0][j+pres_loc]; break;
      case '-':nbase[i][j+pres_loc]=4; break;
      case '?':nbase[i][j+pres_loc]=15; break;
      default: break;

      }
	  
    }
  }
}


/*** Input the number of sequences (nseq), number of sites (nsite)   ****/
/*** and DNA sequences of taxa from datafile.                        ****/
/*** Then transform DNA sequences into numerical sequences (ppBase)  ****/

naym* InputInit(FILESPEC *pFileSpec) {
  
  int i, k;
  char **ppDNASeq;
  int charLen;
  
  charLen=(nsite+1)*sizeof(char);
  ppDNASeq=(char**)malloc(nseq*sizeof(char*));
  if(ppDNASeq==NULL){
    return NULL;
  }
  
  for (i=0; i<nseq; i++)
    {
      ppDNASeq[i]=(char*)malloc(charLen);
      if(ppDNASeq[i]==NULL){
	return NULL;
      }
    }
  
  psname = (naym*)malloc(nseq*sizeof(naym));
  
  if(psname==NULL){
    return NULL;
  }
  
  for (i=0; i<nseq; i++)
    {
      fscanf(pFileSpec->infile, "%s %s", psname[i],ppDNASeq[i]);

      for (k=0; k<11; k++) if (psname[i][k] == 0) psname[i][k] = ' ';

    }
  
  transfer(ppDNASeq,ppBase,0,nsite);
  
  fclose(pFileSpec->infile);
  pFileSpec->infile=NULL;
  
  for (i=0; i<nseq; i++) {
    free(ppDNASeq[i]);
  }
  
  free(ppDNASeq);
  
  return psname;
  
}


/*** Search all unique sites and count them - counts  are   ***/
/*** stored in the array site_counter.                      ***/


/*** This version also eliminates sites with ambiguity codes ***/

void unique_sites(FILESPEC fspec){

  int i,j,k,s,m=0;
  int num_no_gaps=0;
  int start;

  start=0;

  /* Find the first site with no gaps if include_gaps==0 */

  if (include_gaps==0) {

    for (j=0; j<nsite; j++) {      
      i=0;
      /*while (i<nseq && ppBase[i][j]!=4) { i++; }*/
      while (i<nseq && ppBase[i][j]!=4 && ppBase[i][j]!=15) { i++; }
      
      if (i==nseq) {
	start=j;
	break;
      }
      
    }

  }

  /* First remove sites with all gaps if include_gaps==1 */

  if (include_gaps==1) {

    for (j=0; j<nsite; j++) {
      i=0; 
      while (i<nseq && ppBase[i][j]==4) { i++; }
      if (i==nseq) ppBase[0][j] = 99;
    }
  }

  for (j=start; j<nsite; j++) {
    if ((ppBase[0][j]!=99 && include_gaps==1) || (ppBase[0][j]!=99 && include_gaps==0 && ppBase[0][j]!=4 && ppBase[0][j]!=15)) {
      site_counter[m]=1;
      for (i=0; i<nseq; i++) {
	ppBase_unique[i][m] = ppBase[i][j];
      }
      ppBase[0][j]=99;

      for (k=j+1; k<nsite; k++) {
	i=0;
	while (i< nseq && ppBase_unique[i][m] == ppBase[i][k]) {
	  i++;
	}

	if (i==nseq) {
	  site_counter[m]++;
	  ppBase[0][k]=99;
	}

	if (i<nseq && ppBase[i][k]>=4 && include_gaps==0) {
	  ppBase[0][k]=99;
	}

	if (i<nseq && include_gaps==0) {
	  for (s=0; s<nseq; s++) {
	    if (ppBase[s][k]>=4) ppBase[0][k]=99;
	  }
	}
      }
      m++;
    }
  }

  num_unique=m;

  for (i=0; i<num_unique; i++) num_no_gaps += site_counter[i];

}



/*************   The Main Program  *********************/


int main(){

  FILESPEC filespec;
  naym *sname;
  int i, ii, jj, kk, rr, ss, tt, uu;
  int all_quarts, num_quarts;
  float scorej, min_ind, min_score;
  FILE *myinfile;

  /* initialize random generator */
  srand ( time(NULL) );

  /* Keep gaps */
  include_gaps=1;
  
  /*** Read data from files ***/
  if (!FileSpecInit(&filespec)) {
    printf("Can't open needed file \n");
    exit(1);
  }
  fscanf(filespec.infile, "%d %d %d %d", &nseq, &nsite, &all_quarts, &num_quarts);

  printf("There are %d sequences and %d sites\n",nseq,nsite);

  /*** Allocate memory ***/
  MemAlloc();

  /*** Call the function InputInit which will read sequences  ***/
  /*** from infile and convert them to numeric using the      ***/
  /*** function transform.  Returns NULL if a problem occurs. ***/
  sname = InputInit(&filespec);
  if (sname==NULL) {
    printf("Can't allocate memory in InputInit\n");
    exit(1);
  }

  /* Get unique sites */
  unique_sites(filespec);
  fflush(0);

  if (all_quarts==1) { printf("generating all quartets\n");
    for (rr=0; rr<nseq; rr++) {
      for (ss=rr+1; ss<nseq; ss++) {
	for (tt=ss+1; tt<nseq; tt++) {
	  for (uu=tt+1; uu<nseq; uu++) {
	    system("/bin/rm -f infile");
	    myinfile = fopen("infile","w");
	    fprintf(myinfile,"4 %d\n",nsite);

	    fflush(0);

	    // sequence rr
	     if (rr<9) 
	     fprintf(myinfile,"0%d  ",rr+1);
	      else 
	    fprintf(myinfile,"%d  ",rr+1);
	    fflush(0);
	    for (jj=0; jj<num_unique; jj++) {
	      for (ii=0; ii<site_counter[jj]; ii++) {
		if (ppBase_unique[rr][jj]==0) fprintf(myinfile,"A");
		if (ppBase_unique[rr][jj]==1) fprintf(myinfile,"G");
		if (ppBase_unique[rr][jj]==2) fprintf(myinfile,"C");
		if (ppBase_unique[rr][jj]==3) fprintf(myinfile,"T");
		if (ppBase_unique[rr][jj]==4) fprintf(myinfile,"-");
		if (ppBase_unique[rr][jj]==5) fprintf(myinfile,"M");
		if (ppBase_unique[rr][jj]==6) fprintf(myinfile,"R");
		if (ppBase_unique[rr][jj]==7) fprintf(myinfile,"W");
		if (ppBase_unique[rr][jj]==8) fprintf(myinfile,"S");
		if (ppBase_unique[rr][jj]==9) fprintf(myinfile,"Y");
		if (ppBase_unique[rr][jj]==10) fprintf(myinfile,"K");
		if (ppBase_unique[rr][jj]==11) fprintf(myinfile,"B");
		if (ppBase_unique[rr][jj]==12) fprintf(myinfile,"D");
		if (ppBase_unique[rr][jj]==13) fprintf(myinfile,"H");
		if (ppBase_unique[rr][jj]==14) fprintf(myinfile,"V");
		if (ppBase_unique[rr][jj]==15) fprintf(myinfile,"N");		
	      }
	    }
	    fprintf(myinfile,"\n");
	      // sequence ss
	    if (ss<9) 
	    fprintf(myinfile,"0%d  ",ss+1);
	    else 
	    fprintf(myinfile,"%d  ",ss+1);
	      for (jj=0; jj<num_unique; jj++) {
		for (ii=0; ii<site_counter[jj]; ii++) {
		  if (ppBase_unique[ss][jj]==0) fprintf(myinfile,"A");
		  if (ppBase_unique[ss][jj]==1) fprintf(myinfile,"G");
		  if (ppBase_unique[ss][jj]==2) fprintf(myinfile,"C");
		  if (ppBase_unique[ss][jj]==3) fprintf(myinfile,"T");
		  if (ppBase_unique[ss][jj]==4) fprintf(myinfile,"-");
		  if (ppBase_unique[ss][jj]==5) fprintf(myinfile,"M");
		  if (ppBase_unique[ss][jj]==6) fprintf(myinfile,"R");
		  if (ppBase_unique[ss][jj]==7) fprintf(myinfile,"W");
		  if (ppBase_unique[ss][jj]==8) fprintf(myinfile,"S");
		  if (ppBase_unique[ss][jj]==9) fprintf(myinfile,"Y");
		  if (ppBase_unique[ss][jj]==10) fprintf(myinfile,"K");
		  if (ppBase_unique[ss][jj]==11) fprintf(myinfile,"B");
		  if (ppBase_unique[ss][jj]==12) fprintf(myinfile,"D");
		  if (ppBase_unique[ss][jj]==13) fprintf(myinfile,"H");
		  if (ppBase_unique[ss][jj]==14) fprintf(myinfile,"V");
		  if (ppBase_unique[ss][jj]==15) fprintf(myinfile,"N");
		}
	      }
	      fprintf(myinfile,"\n");
	      // sequence tt
	        if (tt<9) 
	      fprintf(myinfile,"0%d  ",tt+1);
	      else 
	      fprintf(myinfile,"%d  ",tt+1);
	      for (jj=0; jj<num_unique; jj++) {
		for (ii=0; ii<site_counter[jj]; ii++) {
		  if (ppBase_unique[tt][jj]==0) fprintf(myinfile,"A");
		  if (ppBase_unique[tt][jj]==1) fprintf(myinfile,"G");
		  if (ppBase_unique[tt][jj]==2) fprintf(myinfile,"C");
		  if (ppBase_unique[tt][jj]==3) fprintf(myinfile,"T");
		  if (ppBase_unique[tt][jj]==4) fprintf(myinfile,"-");
		  if (ppBase_unique[tt][jj]==5) fprintf(myinfile,"M");
		  if (ppBase_unique[tt][jj]==6) fprintf(myinfile,"R");
		  if (ppBase_unique[tt][jj]==7) fprintf(myinfile,"W");
		  if (ppBase_unique[tt][jj]==8) fprintf(myinfile,"S");
		  if (ppBase_unique[tt][jj]==9) fprintf(myinfile,"Y");
		  if (ppBase_unique[tt][jj]==10) fprintf(myinfile,"K");
		  if (ppBase_unique[tt][jj]==11) fprintf(myinfile,"B");
		  if (ppBase_unique[tt][jj]==12) fprintf(myinfile,"D");
		  if (ppBase_unique[tt][jj]==13) fprintf(myinfile,"H");
		  if (ppBase_unique[tt][jj]==14) fprintf(myinfile,"V");
		  if (ppBase_unique[tt][jj]==15) fprintf(myinfile,"N");
		}
	      }
	      fprintf(myinfile,"\n");
	      // sequence uu
	      if (uu<9) 
	      fprintf(myinfile,"0%d ",uu+1);
	      else 
	      fprintf(myinfile,"%d  ",uu+1);
	      for (jj=0; jj<num_unique; jj++) {
		for (ii=0; ii<site_counter[jj]; ii++) {
		  if (ppBase_unique[uu][jj]==0) fprintf(myinfile,"A");
		  if (ppBase_unique[uu][jj]==1) fprintf(myinfile,"G");
		  if (ppBase_unique[uu][jj]==2) fprintf(myinfile,"C");
		  if (ppBase_unique[uu][jj]==3) fprintf(myinfile,"T");
		  if (ppBase_unique[uu][jj]==4) fprintf(myinfile,"-");
		  if (ppBase_unique[uu][jj]==5) fprintf(myinfile,"M");
		  if (ppBase_unique[uu][jj]==6) fprintf(myinfile,"R");
		  if (ppBase_unique[uu][jj]==7) fprintf(myinfile,"W");
		  if (ppBase_unique[uu][jj]==8) fprintf(myinfile,"S");
		  if (ppBase_unique[uu][jj]==9) fprintf(myinfile,"Y");
		  if (ppBase_unique[uu][jj]==10) fprintf(myinfile,"K");
		  if (ppBase_unique[uu][jj]==11) fprintf(myinfile,"B");
		  if (ppBase_unique[uu][jj]==12) fprintf(myinfile,"D");
		  if (ppBase_unique[uu][jj]==13) fprintf(myinfile,"H");
		  if (ppBase_unique[uu][jj]==14) fprintf(myinfile,"V");
		  if (ppBase_unique[uu][jj]==15) fprintf(myinfile,"N");
		}
	      }
	      fprintf(myinfile,"\n");
	      fclose(myinfile);

	      system("./SVDscores > temp.out");
	      system("cat quartets.out temp.out > newtemp");
	      system("/bin/cp -f newtemp quartets.out");
	      //exit(1);
	  }
	}
      }
    }
    

  }/*end if (all_quarts==1)*/

  else {

    printf("Randomly sampling %d quartets\n",num_quarts);
    for (kk=0; kk<num_quarts; kk++) {

      rr=ss=tt=uu=1;

      /* sample 4 taxa to include */
      rr = rand()%nseq;
      while (ss==rr) ss = rand()%nseq;
      while ((tt==rr) || (tt==ss)) tt = rand()%nseq;
      while ((uu==rr) || (uu==ss) || (uu==tt)) uu = rand()%nseq;

      system("/bin/rm -f infile");
      myinfile = fopen("infile","w");
      fprintf(myinfile,"4 %d\n",nsite);

      if (rr<9) 
      fprintf(myinfile,"0%d  ",rr+1);
      else 
      fprintf(myinfile,"%d  ",rr+1);
      fflush(0);
      for (jj=0; jj<num_unique; jj++) {
	for (ii=0; ii<site_counter[jj]; ii++) {
	  if (ppBase_unique[rr][jj]==0) fprintf(myinfile,"A");
	  if (ppBase_unique[rr][jj]==1) fprintf(myinfile,"G");
	  if (ppBase_unique[rr][jj]==2) fprintf(myinfile,"C");
	  if (ppBase_unique[rr][jj]==3) fprintf(myinfile,"T");
	  if (ppBase_unique[rr][jj]==4) fprintf(myinfile,"-");
	  if (ppBase_unique[rr][jj]==5) fprintf(myinfile,"M");
	  if (ppBase_unique[rr][jj]==6) fprintf(myinfile,"R");
	  if (ppBase_unique[rr][jj]==7) fprintf(myinfile,"W");
	  if (ppBase_unique[rr][jj]==8) fprintf(myinfile,"S");
	  if (ppBase_unique[rr][jj]==9) fprintf(myinfile,"Y");
	  if (ppBase_unique[rr][jj]==10) fprintf(myinfile,"K");
	  if (ppBase_unique[rr][jj]==11) fprintf(myinfile,"B");
	  if (ppBase_unique[rr][jj]==12) fprintf(myinfile,"D");
	  if (ppBase_unique[rr][jj]==13) fprintf(myinfile,"H");
	  if (ppBase_unique[rr][jj]==14) fprintf(myinfile,"V");
	  if (ppBase_unique[rr][jj]==15) fprintf(myinfile,"N");
	}
      }
      fprintf(myinfile,"\n");
      // sequence ss
      if (ss<9) 
      fprintf(myinfile,"0%d  ",ss+1);
      else 
      fprintf(myinfile,"%d  ",ss+1);
      for (jj=0; jj<num_unique; jj++) {
	for (ii=0; ii<site_counter[jj]; ii++) {
	  if (ppBase_unique[ss][jj]==0) fprintf(myinfile,"A");
	  if (ppBase_unique[ss][jj]==1) fprintf(myinfile,"G");
	  if (ppBase_unique[ss][jj]==2) fprintf(myinfile,"C");
	  if (ppBase_unique[ss][jj]==3) fprintf(myinfile,"T");
	  if (ppBase_unique[ss][jj]==4) fprintf(myinfile,"-");
	  if (ppBase_unique[ss][jj]==5) fprintf(myinfile,"M");
	  if (ppBase_unique[ss][jj]==6) fprintf(myinfile,"R");
	  if (ppBase_unique[ss][jj]==7) fprintf(myinfile,"W");
	  if (ppBase_unique[ss][jj]==8) fprintf(myinfile,"S");
	  if (ppBase_unique[ss][jj]==9) fprintf(myinfile,"Y");
	  if (ppBase_unique[ss][jj]==10) fprintf(myinfile,"K");
	  if (ppBase_unique[ss][jj]==11) fprintf(myinfile,"B");
	  if (ppBase_unique[ss][jj]==12) fprintf(myinfile,"D");
	  if (ppBase_unique[ss][jj]==13) fprintf(myinfile,"H");
	  if (ppBase_unique[ss][jj]==14) fprintf(myinfile,"V");
	  if (ppBase_unique[ss][jj]==15) fprintf(myinfile,"N");
	}
      }
      fprintf(myinfile,"\n");
      // sequence tt
      if (tt<9) 
      	fprintf(myinfile,"0%d  ",tt+1);
      else 
      fprintf(myinfile,"%d  ",tt+1);
      for (jj=0; jj<num_unique; jj++) {
	for (ii=0; ii<site_counter[jj]; ii++) {
	  if (ppBase_unique[tt][jj]==0) fprintf(myinfile,"A");
	  if (ppBase_unique[tt][jj]==1) fprintf(myinfile,"G");
	  if (ppBase_unique[tt][jj]==2) fprintf(myinfile,"C");
	  if (ppBase_unique[tt][jj]==3) fprintf(myinfile,"T");
	  if (ppBase_unique[tt][jj]==4) fprintf(myinfile,"-");
	  if (ppBase_unique[tt][jj]==5) fprintf(myinfile,"M");
	  if (ppBase_unique[tt][jj]==6) fprintf(myinfile,"R");
	  if (ppBase_unique[tt][jj]==7) fprintf(myinfile,"W");
	  if (ppBase_unique[tt][jj]==8) fprintf(myinfile,"S");
	  if (ppBase_unique[tt][jj]==9) fprintf(myinfile,"Y");
	  if (ppBase_unique[tt][jj]==10) fprintf(myinfile,"K");
	  if (ppBase_unique[tt][jj]==11) fprintf(myinfile,"B");
	  if (ppBase_unique[tt][jj]==12) fprintf(myinfile,"D");
	  if (ppBase_unique[tt][jj]==13) fprintf(myinfile,"H");
	  if (ppBase_unique[tt][jj]==14) fprintf(myinfile,"V");
	  if (ppBase_unique[tt][jj]==15) fprintf(myinfile,"N");
	}
      }
      fprintf(myinfile,"\n");
      // sequence uu
      if (uu<9) 
      fprintf(myinfile,"0%d ",uu+1);
      else 
      fprintf(myinfile,"%d  ",uu+1);
      for (jj=0; jj<num_unique; jj++) {
	for (ii=0; ii<site_counter[jj]; ii++) {
	  if (ppBase_unique[uu][jj]==0) fprintf(myinfile,"A");
	  if (ppBase_unique[uu][jj]==1) fprintf(myinfile,"G");
	  if (ppBase_unique[uu][jj]==2) fprintf(myinfile,"C");
	  if (ppBase_unique[uu][jj]==3) fprintf(myinfile,"T");
	  if (ppBase_unique[uu][jj]==4) fprintf(myinfile,"-");
	  if (ppBase_unique[uu][jj]==5) fprintf(myinfile,"M");
	  if (ppBase_unique[uu][jj]==6) fprintf(myinfile,"R");
	  if (ppBase_unique[uu][jj]==7) fprintf(myinfile,"W");
	  if (ppBase_unique[uu][jj]==8) fprintf(myinfile,"S");
	  if (ppBase_unique[uu][jj]==9) fprintf(myinfile,"Y");
	  if (ppBase_unique[uu][jj]==10) fprintf(myinfile,"K");
	  if (ppBase_unique[uu][jj]==11) fprintf(myinfile,"B");
	  if (ppBase_unique[uu][jj]==12) fprintf(myinfile,"D");
	  if (ppBase_unique[uu][jj]==13) fprintf(myinfile,"H");
	  if (ppBase_unique[uu][jj]==14) fprintf(myinfile,"V");
	  if (ppBase_unique[uu][jj]==15) fprintf(myinfile,"N");
	}
      }
      fprintf(myinfile,"\n");
      fclose(myinfile);

      system("./SVDscores > temp.out");
      system("cat quartets.out temp.out > newtemp");
      system("/bin/cp -f newtemp quartets.out");


    }

  } /*end else (all_quarts==1) */
   
}


  



